#pragma once


#include <vector>
#include <string>
#include "../../tinyxml/tinyxml/tinyxml.h"
#include "../../system/misc/misc.h"
#include "../../system/crtsec/crtsec.h"


namespace state_machine
{


class StateMachine
{
public:
	class State
	{
	public:
		static int id_counter;
	public:
		struct transition_t
		{
			std::string input;
			std::string output;
			const State *next;

			bool on_manual;

		public:
			transition_t();
			transition_t( const transition_t &src );
			~transition_t();

			void print( const char *white_spaces = "" ) const;
		};

	private:
		int id;
		std::string name;
		std::vector< transition_t > trans_tbl;
		
		const State *error;
		
		std::vector< std::string > killed_on_error;

	public:
		State();
		State( const State &src );
		~State();

	public:
		bool initialize( int id, const std::string &name );
		bool setTransitionTable( const std::vector< transition_t > &trans_tbl );
		bool setRTCsKilledOnError( const std::vector< std::string > &kill_list );
		bool finalize();

	public:
		int getID() const { return id; }

		const std::string &getName() const { return name; }
		std::string &getName() { return name; }

		const std::vector< transition_t > &getTransitionTable() const { return trans_tbl; }
		std::vector< transition_t > &getTransitionTable() { return trans_tbl; }

		int getRequiredInput( std::string &required_input ) const;
		int getRequiredInput( const std::string &next_state, std::string &required_input ) const;

		const State *getError() const { return error;	}
		void setError( const State *state ) { error = state;	}

		//bool killRTCs();
		
		const std::vector< std::string > &getRTCsKilledOnError() const
		{ return killed_on_error;  }
		std::vector< std::string > &getRTCsKilledOnError()
		{ return killed_on_error;  }

		void print() const;
	};


	class ScenarioManager
	{
	private:
		bool is_initialized;
		std::vector< std::string > rtcs;
		std::vector< std::string > nameservers;

	public:
		ScenarioManager();
		~ScenarioManager();

	public:
		// (XML̃[h)
		int initialize( const char *xml_path, std::vector< State > &state_tbl );
		int finalize();
		
		void print();

	public:
		const std::vector< std::string > &getRTCs() const { return rtcs; }
		const std::vector< std::string > &getNameServers() const { return nameservers; }
		bool isInitialized() const { return is_initialized; }

	private:	// [_[
		// <rtm>
		int loadRTM( TiXmlDocument &scenario );
		// <nameservers>
		int loadNameServersInRTM( TiXmlNode *rtm_node );
		// <rtcs>
		int loadRTCsIInRTM( TiXmlNode *rtm_node );
		// <statemachine>
		int loadStateMachine( TiXmlDocument &scenario, std::vector< State > &state_tbl );
		// <state>
		int loadStatesInStateMachine( TiXmlNode *state_machine_node, std::vector< State > &state_tbl );
		int loadStateInStateMachine( TiXmlNode *state_node, std::vector< State > &state_tbl, State &state );
		int loadNextInState( TiXmlNode *next_state_node, std::vector< State > &state_tbl, State::transition_t &trans );
		int loadErrorInState( TiXmlNode *error_node, std::vector< State > &state_tbl, State &error );
		int loadRTCKilledInState( TiXmlNode *kill_node, State &state );

		int loadValue( TiXmlNode *node, const char *node_name, std::string &value, const char *missing = NULL );
		int loadValue( TiXmlNode *node, std::string &value, const char *missing = NULL );
	private:
		// G[bZ[W̕\
		template <typename T>
		T returnByNotFound(
			T code, const char *node_or_element, bool wait_for_return = false )
		{
			static std::string msg;
			msg = std::string( node_or_element );
			msg += std::string( ": not found." );
			return sys::returnByError< T >( code, msg.c_str(), wait_for_return );
		}
		template <typename T>
		T returnByUndefined(
			T code, const char *node_or_element, bool wait_for_return = false )
		{
			static std::string msg;
			msg = std::string( node_or_element );
			msg += std::string( ": undefined node." );
			return sys::returnByError< T >( code, msg.c_str(), wait_for_return );
		}

	};
	

private:
	bool in_transition;

	std::vector< State > state_tbl;
	const State *curr;
	ScenarioManager scn_mgr;
	sys::CriticalSection cs;

public:
	StateMachine();
	~StateMachine();

public:
	int initialize( const char *scenario_path );
	int finalize();
	
	void print();

public:
	int transition( const char *input );
	int transition( const char *input, std::string &output );
	int transition( const std::string &input )
	{ return transition( input.c_str() ); }


	int setCurrentState( int id );
	int setCurrentState( const char *name );

public:
	const State &getCurrentState() const { return *curr; }

	// Ԃ̏oɑ΂ANZbT
	const std::string &getCurrentStateName() const { return curr->getName(); }
	int isNextStateFound( const char *input );
	int isNextStateFound( const std::string &input )
	{	return isNextStateFound( input.c_str() );	}

	int getRequiredInput( const std::string &next_state, std::string &input ) const
	{ return curr->getRequiredInput( next_state, input ); }

	const std::vector< State::transition_t > &getTransitionTable() const
	{ return curr->getTransitionTable(); }

	int getError( std::string &error_state_name ) const;
	int transitionIntoError();

	const std::vector< std::string > &getNameServers() const
	{ return scn_mgr.getNameServers(); }

	// NeBJZNV
	void lock() { cs.enter(); }
	void unlock() { cs.exit(); }

};



}